# $Id: aws.rb,v 1.23 2008/04/12 14:08:32 ianmacd Exp $
#

module Amazon
  module AWS

    require 'uri'
    require 'amazon'
    require 'amazon/aws/cache'
    require 'rexml/document'

    NAME = '%s/%s' % [ Amazon::NAME, 'AWS' ]
    VERSION = '0.1.0'
    USER_AGENT = '%s %s' % [ NAME, VERSION ]

    # Default Associate tags to use per locale.
    #
    DEF_ASSOC  = {
      'ca' => 'caliban-20',
      'de' => 'calibanorg0a-21',
      'fr' => 'caliban08-21',
      'jp' => 'calibanorg-20',
      'uk' => 'caliban-21',
      'us' => 'calibanorg-20'
    }

    # Service name and version for AWS.
    #
    SERVICE = { 'Service' => 'AWSECommerceService',
		'Version' => '2008-03-03'
    }

    # Maximum number of 301 and 302 HTTP responses to follow, should Amazon
    # later decide to change the location of the service.
    #
    MAX_REDIRECTS = 3

    # Maximum number of results pages that can be retrieved for a given
    # search.
    #
    MAX_PAGES = 400


    class Endpoint

      attr_reader :host, :path

      def initialize(endpoint)
	uri = URI.parse( endpoint )
	@host = uri.host
	@path = uri.path
      end
    end

    ENDPOINT = {
      'ca' => Endpoint.new( 'http://ecs.amazonaws.ca/onca/xml' ),
      'de' => Endpoint.new( 'http://ecs.amazonaws.de/onca/xml' ),
      'fr' => Endpoint.new( 'http://ecs.amazonaws.fr/onca/xml' ),
      'jp' => Endpoint.new( 'http://ecs.amazonaws.jp/onca/xml' ),
      'uk' => Endpoint.new( 'http://ecs.amazonaws.co.uk/onca/xml' ),
      'us' => Endpoint.new( 'http://ecs.amazonaws.com/onca/xml' )
    }

    # Fetch a page, either from the cache or by HTTP. This is used internally.
    #
    def AWS.get_page(request, query)

      url = ENDPOINT[request.locale].path + query
      cache_url = ENDPOINT[request.locale].host + url

      # Check for cached page and return that if it's there.
      #
      if request.cache && request.cache.cached?( cache_url )
	body = request.cache.get_cached( cache_url )
	return body if body
      end

      # Get the existing connection. If there isn't one, force a new one.
      #
      conn = request.conn || request.reconnect.conn
      user_agent = request.user_agent

      Amazon::dprintf( 'Fetching http://%s%s ...', conn.address, url )

      begin
	response = conn.get( url, { 'user-agent', user_agent } )

      # If we've pulled and processed a lot of pages from the cache (or
      # just not passed by here recently), the HTTP connection to the server
      # will probably have timed out.
      #
      rescue Errno::ECONNRESET
	conn = request.reconnect.conn
	retry
      end

      redirects = 0
      while response.key? 'location'
	if ( redirects += 1 ) > MAX_REDIRECTS
	  raise HTTPError, "More than #{MAX_REDIRECTS} redirections"
	end

	old_url = url
	url = URI.parse( response['location'] )
	url.scheme = old_url.scheme unless url.scheme
	url.host = old_url.host unless url.host
	Amazon::dprintf( 'Following HTTP %s to %s ...', response.code, url )
	response = Net::HTTP::start( url.host ).
      	       get( url.path,{ 'user-agent', user_agent } )
      end

      if response.code != '200'
	raise HTTPError, "HTTP response code #{response.code}"
      end

      # Cache the page if we're using a cache.
      #
      if request.cache
	request.cache.cache( cache_url, response.body )
      end

      response.body
    end


    def AWS.assemble_query(items)
      query = ''

      # We must sort the items into an array to get reproducible ordering
      # of the query parameters. Otherwise, URL caching would not work.
      #
      items.sort.each do |k, v|
	query << '&%s=%s' % [ k, Amazon.url_encode( v.to_s ) ]
      end

      # Replace initial ampersand with question-mark.
      #
      query[0] = '?'

      query
    end


    # Everything returned by AWS is an AWSObject.
    #
    class AWSObject

      include REXML

      def initialize
	# The name of this instance variable must never clash with the
	# uncamelised name of an Amazon tag.
	#
	# This is used to store the REXML::Text value of an element, which
	# exists only when the element contains no children.
	#
	@__val__ = nil
      end

      def method_missing(method, *params)
	iv = '@' + method.id2name
	instance_variables.include?( iv ) ? instance_variable_get( iv ) : nil
      end
      private :method_missing
 

      def remove_val
	remove_instance_variable( :@__val__ )
      end
      private :remove_val


      def inspect
	remove_val if instance_variable_defined?( :@__val__ ) && @__val__.nil?
	str = super
	str.sub( /@__val__=/, 'value=' ) if str
      end


      def to_s
	if instance_variable_defined?( :@__val__ )
	  return @__val__ if @__val__.is_a?( String )
	  remove_val 
	end

	string = ''

	# Get all attributes, minus the leading '@'.
	#
	vars = self.instance_variables.sort.map { |v| v[1..-1] }

	# Find the longest attribute name.
	#
	longest = vars.sort { |a,b| b.length <=> a.length }[0].length

	# Display the object's details.
	#
	vars.each do |iv|
	  value = self.instance_variable_get( "@#{iv}" ).inspect
	  string << "%-#{longest}s = %s\n" % [ iv, value ]
	end

	string
      end


      # This alias makes the ability to determine an AWSObject's properties a
      # little more intuitive. It's pretty much just an alias for the
      # inherited <em>Object#instance_variables</em> method.
      #
      def properties
	iv = instance_variables
	iv.delete( '@__val__' )
	iv
      end


      # Convert an AWSObject to a Hash.
      #
      def to_h
	hash = {}
	self.properties.each do |property|
	  key = property[1..-1]
	  val = self.instance_variable_get( property )
	  hash[key] = val
	end

	hash
      end


      # Fake the appearance of an AWSObject as a hash. _key_ should be any
      # attribute of the object and can be a String, Symbol or anything else
      # that can be converted to a String with to_s.
      #
      def [](key)
	self.instance_variable_get( '@' + key.to_s )
      end


      # Recursively walk through an XML tree, starting from _node_. This is
      # called internally and is not intended for user code.
      #
      def walk(node)
    
	if node.instance_of?( REXML::Document )
	  walk( node.root )
    
	elsif node.instance_of?( REXML::Element )
	  name = Amazon.uncamelise( node.name )
    
	  # Create a class for the new element type unless it already exists.
	  #
	  unless AWS::AWSObject.const_defined?( node.name )
	    cl = AWS::AWSObject.const_set( node.name, Class.new( AWSObject ) )

	    # Give it an accessor for @attrib.
	    #
	    cl.send( :attr_accessor, :attrib )
	  end
    
	  # Instantiate an object in the newly created class.
	  #
	  obj = AWS::AWSObject.const_get( node.name ).new

	  sym_name = "@#{name}".to_sym
    
	  if instance_variable_defined?( sym_name)
    	    instance_variable_set( sym_name,
    	      instance_variable_get( sym_name ) << obj )
	  else
	    instance_variable_set( sym_name, AWSArray.new( [ obj ] ) )
	  end
    
	  if node.has_attributes?
	    obj.attrib = {}
	    node.attributes.each_pair do |name, value|
	      obj.attrib[name.downcase] = value.to_s.sub( /^#{name}=/, '' )
	    end
	  end

	  node.children.each { |child| obj.walk( child ) }
    
	else # REXML::Text
	  @__val__ = node.to_s
	end
      end

    end


    # Everything we get back from AWS is transformed into an array. Many of
    # these, however, have only one element (because the corresponding XML
    # consists of a parent element containing only a single child element).
    #
    # This class consists solely to allow single element arrays to pass a
    # method call down to their one element, thus obviating the need for lots
    # of references to <tt>foo[0]</tt> in user code.
    #
    # For example, the following:
    #
    #  items = resp.item_search_response[0].items[0].item
    #
    # can be reduced to:
    #
    #  items = resp.item_search_response.items.item
    #
    class AWSArray < Array

      def method_missing(method, *params)
	self.size == 1 ? self[0].send( method, *params ) : super
      end
      private :method_missing

    end

 
    # This is the base class of all AWS operations.
    #
    class Operation

      # These are the types of search currently implemented by Ruby/AWS.
      #
      OPERATIONS = %w[
	BrowseNodeLookup ItemLookup ItemSearch ListSearch SellerListingSearch
      ]

      # These are the valid search parameters that can be used with
      # ItemSearch.
      #
      PARAMETERS = %w[
	Actor		Artist	      AudienceRating	Author
	Brand		BrowseNode    City Composer	Conductor
	Director	Keywords      Manufacturer	MusicLabel
	Neighborhood	Orchestra     Power		Publisher
	TextStream	Title
      ]

      OPT_PARAMETERS = %w[
	Availability Condition MinimumPrice MaximumPrice
	MerchantId OfferStatus Sort
      ]

      ALL_PARAMETERS = PARAMETERS + OPT_PARAMETERS

      attr_reader :kind
      attr_accessor :params

      def initialize(parameters)

	op_kind = self.class.to_s.sub( /^.*::/, '' )
	unless OPERATIONS.include?( op_kind ) || op_kind == 'MultipleOperation'
	  raise "Bad operation: #{op_kind}"
	end
	#raise 'Too many parameters' if parameters.size > 10

	@kind = op_kind
	@params = { 'Operation' => op_kind }.merge( parameters )
      end


      # Convert parameters to batch format, e.g. ItemSearch.1.Title.
      #
      def batch_parameters(params, *b_params)

	@index ||= 1

	unless b_params.empty?
	  class_str = self.class.to_s.sub( /^.+::/, '' )
	  all_parameters = [ params ].concat( b_params )
	  params = {}

	  all_parameters.each_with_index do |hash, index|

	    # Don't batch an already batched hash.
	    #
	    if ! hash.empty? && hash.to_a[0][0] =~ /^.+\..+\..+$/
	      params = hash
	      next
	    end

	    hash.each do |tag, val|
	      shared_param = '%s.%d.%s' % [ class_str, @index + index, tag ]
	      params[shared_param] = val
	    end
	  end

	  @index += b_params.size

	end

	params
      end


      def parameter_check(parameters)
	parameters.each_key do |key|
	  raise "Bad parameter: #{key}" unless ALL_PARAMETERS.include? key
	end
      end
      private :parameter_check

    end


    # This class can be used to merge operations into a single operation.
    # AWS currently supports combining two operations, 
    #
    class MultipleOperation < Operation

      # This will allow you to take two Operation objects and combine them to
      # form a single object, which can then be used to perform searches. AWS
      # itself imposes the maximum of two combined operations.
      #
      # <em>operation1</em> and <em>operation2</em> are both objects from a
      # subclass of Operation, such as ItemSearch, ItemLookup, etc.
      #
      # There are currently a few restrictions in the Ruby/AWS implementation
      # of multiple operations:
      #
      # - ResponseGroup objects used when calling AWS::Search::Request#search
      #   apply to both operations. You cannot have a separate ResponseGroup
      #   set per operation.
      #
      # - One or both operations may have multiple results pages available,
      #   but only the first page can be returned. If you need the other
      #   pages, perform the operations separately, not as part of a
      #   MultipleOperation.
      #
      # Example:
      #
      #  is = ItemSearch.new( 'Books', { 'Title' => 'Ruby' } )
      #  il = ItemLookup.new( 'ASIN', { 'ItemId' => 'B0013DZAYO',
      #					'MerchantId' => 'Amazon' } )
      #	 mo = MultipleOperation.new( is, il )
      # 
      def initialize(operation1, operation2)

	# Safeguard against changing original Operation objects in place. This
	# is to protect me, not for user code.
	#
	operation1.freeze
	operation2.freeze

	op_kind = '%s,%s' % [ operation1.kind, operation2.kind ]

	# Duplicate Operation objects and remove their Operation parameter.
	# 
	op1 = operation1.dup
	op1.params = op1.params.dup
	op1.params.delete( 'Operation' )

	op2 = operation2.dup
	op2.params = op2.params.dup
	op2.params.delete( 'Operation' )

	if op1.class == op2.class

	  # If both operations are of the same type, we combine the parameters
	  # of both.
	  #
	  b_params = op1.batch_parameters( op1.params, op2.params )
	else

	  # We have to convert the parameters to batch format.
	  #
	  bp1 = op1.batch_parameters( op1.params, {} )
	  bp2 = op2.batch_parameters( op2.params, {} )
	  b_params = bp1.merge( bp2 )
	end

	params = { 'Operation' => op_kind }.merge( b_params )
	super( params )

      end
      
    end


    # This is the class for the most common type of AWS look-up, an
    # ItemSearch. This allows you to search for items that match a set of
    # broad criteria.
    #
    class ItemSearch < Operation

      # Not all search indices work in all locales. It is the user's
      # responsibility to ensure that a given index is valid within a given
      # locale.
      #
      # According to the AWS documentation:
      #
      # - *All* searches through all indices.
      # - *Blended* combines DVD, Electronics, Toys, VideoGames, PCHardware,
      #   Tools, SportingGoods, Books, Software, Music, GourmetFood, Kitchen
      #   and Apparel.
      # - *Merchants* combines all search indices for a merchant given with
      #   MerchantId.
      # - *Music* combines the Classical, DigitalMusic, and MusicTracks
      #   indices.
      # - *Video* combines the DVD and VHS search indices.
      #
      SEARCH_INDICES = %w[
	    All
	    Apparel	      Hobbies	      PetSupplies
	    Automotive	   HomeGarden	   Photo
	    Baby		 Jewelry	      Software
	    Beauty	       Kitchen	      SoftwareVideoGames
	    Blended	      Magazines	    SportingGoods
	    Books		Merchants	    Tools
	    Classical	    Miscellaneous	Toys
	    DigitalMusic	 Music		VHS
	    DVD		  MusicalInstruments   Video
	    Electronics	  MusicTracks	  VideoGames
	    ForeignBooks	 OfficeProducts       Wireless
	    GourmetFood	  OutdoorLiving	WirelessAccessories
	    HealthPersonalCare   PCHardware
	]


      # Search AWS for items. _search_index_ must be one of _SEARCH_INDICES_
      # and _parameters_ is a hash of relevant search parameters.
      #
      # Example:
      #
      #  is = ItemSearch.new( 'Books', { 'Title' => 'ruby programming' } )
      #
      def initialize(search_index, parameters)
	unless SEARCH_INDICES.include? search_index
	  raise "Invalid search index: #{search_index}"
	end

	parameter_check( parameters )
	super( { 'SearchIndex' => search_index }.merge( parameters ) )
      end
    end


    # This class of look-up deals with searching for *specific* items by some
    # uniquely identifying attribute, such as the ASIN (Amazon Standard Item
    # Number).
    #
    class ItemLookup < Operation

      # Look up a specific item in the AWS catalogue. _id_type_ is the type of
      # identifier, _parameters_ is a hash that identifies the item to be
      # located and narrows the scope of the search, and _b_parameters_ is an
      # optional hash of further items to be located. Use of _b_parameters_
      # effectively results in a batch operation being sent to AWS.
      #
      # Example:
      #
      #  il = ItemLookup.new( 'ASIN', { 'ItemId' => 'B000AE4QEC'
      #				 'MerchantId' => 'Amazon' },
      #			       { 'ItemId' => 'B000051WBE',
      #				 'MerchantId' => 'Amazon' } )
      #
      # In the above example, we search for two items, based on their ASIN.
      # The use of _MerchantId_ restricts the offers returned to those for
      # sale by Amazon (as opposed to third-party sellers).
      #
      def initialize(id_type, parameters, *b_parameters)

	id_type_str = 'IdType'

	unless b_parameters.empty?
	  class_str = self.class.to_s.sub( /^.+::/, '' )
	  id_type_str = '%s.Shared.IdType' % [ class_str ]
	  parameters = batch_parameters( parameters, *b_parameters )
	end

	super( { id_type_str => id_type }.merge( parameters ) )
      end
    end


    # Search for items for sale by a particular seller.
    #
    class SellerListingSearch < Operation

      # Search for items for sale by a particular seller. _seller_id_ is the
      # Amazon seller ID and _parameters_ is a hash of parameters that narrows
      # the scope of the search.
      #
      # Example:
      #
      #  sls = SellerListingSearch.new( 'A33J388YD2MWJZ',
      #					{ 'Keywords' => 'Killing Joke' } )
      #
      def initialize(seller_id, parameters)
	super( { 'SellerId' => seller_id }.merge( parameters ) )
      end
    end


    # Find wishlists, registry lists, etc. placed on Amazon by users.
    #
    class ListSearch < Operation

      # Search for lists on Amazon. _list_type_ is the type of list to search
      # for and _parameters_ is a hash of parameters that narrows the scope of
      # the search.
      #
      # Example:
      #
      #  ls = ListSearch.new( 'WishList', { 'Name' => 'Frank Cusack' }
      #
      def initialize(list_type, parameters)
	super( { 'ListType' => list_type}.merge( parameters ) )
      end
    end


    # Look up Amazon browse nodes.
    #
    class BrowseNodeLookup < Operation

      # Look up and return the details of an Amazon browse node. _node_ is the
      # browse node to look up and _parameters_ is a hash of parameters that
      # serves to further define the search. _parameters_ is currently unused.
      #
      # Example:
      #
      #  bnl = BrowseNodeLookup.new( '11232', {} )
      #
      def initialize(node, parameters={})
	super( { 'BrowseNodeId' => node }.merge( parameters ) )
      end
    end


    # Response groups determine how much data related to the item(s) being
    # sought is returned.
    #
    class ResponseGroup

      RESPONSE_GROUPS = %w[
	BrowseNodeInfo BrowseNodes Large Medium Small Offers OfferFull
      ]	 # :nodoc:

      attr_reader :list, :params

      # Define a set of one or more response groups to be applied to items
      # found in a search operation.
      #
      # If no response groups are given in _rg_ when instantiating an object,
      # *Small* will be used by default.
      #
      # Example:
      #
      #  rg = ResponseGroup.new( 'Medium', 'Offers', 'Reviews' )
      #
      def initialize(*rg)
	rg << 'Small' if rg.empty?
	@list = rg
	@params = { 'ResponseGroup' => @list.join( ',' ) }
      end
    end


    # All dynamically generated exceptions occur within this namespace.
    #
    module Error

      # An exception generator class.
      #
      class AWSError

	attr_reader :exception

	def initialize(xml)
	  err_class = xml.elements['Code'].text.sub( /^AWS.*\./, '' )
	  err_msg = xml.elements['Message'].text

	  unless Amazon::AWS::Error.const_defined?( err_class )
	    Amazon::AWS::Error.const_set( err_class,
					  Class.new( StandardError ) )
	  end

	  ex_class = Amazon::AWS::Error.const_get( err_class )
	  @exception = ex_class.new( err_msg )
	end

      end

    end

  end

end
